#!/usr/bin/env python

#Copyright 2010, HudsonAlpha Institute for Biotechnology
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

#unzip.py, a program to extract zip archives with large files
#Written by Phillip Dexheimer, June 2010

version_str = "1.0.0"
   
import fnmatch
from optparse import OptionParser
import os
import sys
import zipfile
import zlib #For CRC

def pass_whitelist(filename, whitelist):
    '''Returns True if whitelist is empty or filename matches a
    pattern in the whitelist'''
    if len(whitelist) == 0:
        return True
    for spec in whitelist:
        if fnmatch.fnmatch(filename, spec):
            return True
    return False

def pass_blacklist(filename, blacklist):
    '''Returns True if blacklist is empty or filename does not match
    any patterns in the blacklist'''
    if len(blacklist) == 0:
        return True
    for spec in blacklist:
        if fnmatch.fnmatch(filename, spec):
            return False
    return True

def list_files(zip, infolist):
    '''Given a list of ZipInfo structures, displays information about
    them on stdout'''
    list_comment(zip)
    print "  Length      Date    Time    Name"
    print " --------     ----    ----    ----"
    count = 0
    size = 0
    for info in infolist:
        count += 1
        size += info.file_size
        print "{0:>9}  {1[1]:0=2}-{1[2]:0=2}-{1[0]:4} {1[3]:0=2}:{1[4]:0=2}   {2}".format(
            info.file_size, info.date_time, info.filename)
    print " --------                     -------"
    print "{0:>9}                     {1} files".format(size, count)
    
def test_files(zip, infolist):
    '''Given a list of ZipInfo structures, verifies that the CRC of each one
    matches the stored value'''
    chunksize = 1 << 11 # 2 MB
    is_clean = True
    for info in infolist:
        print "    testing: {0}  ".format(info.filename),
        sys.stdout.flush()
        
        f = zip.open(info)
        data = f.read(chunksize)
        value = zlib.crc32(data)
        while len(data) == chunksize:
            data = f.read(chunksize)
            value = zlib.crc32(data, value)
        f.close()
        
        if (value & 0xffffffff) == info.CRC:
            print "OK"
        else:
            print "FAIL"
            is_clean = False
    if is_clean:
        if len(infolist) == len(zip.infolist()):
            print "No errors detected in compressed data of {0}.".format(zip.filename)
        else:
            print "No errors detected in {0} for the {1} files tested.".format(zip.filename, len(infolist))
    else:
        print "Errors detected in {0}".format(zip.filename)

def list_comment(zip):
    '''Prints out the archive comment'''
    if zip.comment:
        print zip.comment

def get_input(prompt, valid):
    '''Displays prompt to the user, repeating until a valid response is received'''
    while 1:
        result = raw_input(prompt)
        if result in valid:
            return result
        print "error:  invalid response [{0}]".format(result)

def extract(zip, infolist, overwrite):
    '''Extracts the files from the archive zip that are specified in infolist.
    Note that the extracted files will be named according to the provided
    ZipInfo::filename attributes.  overwrite may contain the values 'always',
    'never', or 'prompt' '''
    for info in infolist:
        if os.path.lexists(info.filename):
            if overwrite == "never":
                continue
            if overwrite == "prompt":
                result = get_input("replace {0}? [y]es, [n]o, [A]ll, [N]one, [r]ename: ".format(info.filename),
                                   ("y", "n", "A", "N", "r"))
                if result == 'n':
                    continue
                elif result == 'N':
                    overwrite = "never"
                    continue
                elif result == 'A':
                    overwrite = "always"
                elif result == 'r':
                    newname = raw_input("new name: ")
                    info.filename = os.path.join(os.path.dirname(info.filename), newname)
        if info.compress_type == zipfile.ZIP_STORED:
            print " extracting:",
        else:
            print "  inflating:",
        print info.filename,
        sys.stdout.flush()
        zip.extract(info)
        print

def adjust_path(filename, extract_root, junk_paths):
    '''Given a desired extraction point and whether the user wants to junk the
    directory structure in the archive, returns the appropriately modified path
    to filename'''
    path, name = os.path.split(filename)
    if junk_paths:
        path = ""
    if extract_root:
        path = os.path.join(extract_root, path)
    return os.path.join(path, name)

if __name__ == '__main__':
    usage = """usage: %prog [options] file[.zip] [file(s)...]
    
    file[.zip] is the path to the zip archive(s) to process.  Wildcards are
    permitted, but must be escaped so that they are not processed by the shell.
    Only the filename can be a wildcard, the path itself cannot
    
    The remaining arguments are an (optional) list of files in the archive to
    operate on.  Wildcards are permitted."""
    parser = OptionParser(usage=usage, version="%prog "+version_str)
    parser.add_option("-d", metavar="exdir", dest="root", 
                      help="Directory to extract the archive into (default: current directory)")
    parser.add_option("-x", metavar="xfile", dest="exclude", action="append",
                      default=[],
                      help="Archive members to be excluded from processing.  Wildcards "
                      "are permitted.  Multiple -x options may be specified")
    parser.add_option("-j", action="store_true", dest="junk", default=False,
                      help="Junk paths.  The directory structure of the archive is "
                      "not recreated, all files are placed in the extraction directory")
    parser.add_option("-l", action="store_true", dest="list", default=False,
                      help="List the name, size, and modification date of all files "
                      "specified")
    parser.add_option("-t", action="store_true", dest="test", default=False,
                      help="Test the specified archived members.  Files are extracted "
                      "into memory and their checksum compared to the stored value")
    parser.add_option("-z", action="store_true", dest="display_comment", default=False,
                      help="Display the archive comment and quits")
    parser.add_option("-n", action="store_const", const="never", dest="overwrite",
                      help="Never overwrite files.  Files are not extracted if a file "
                      "with the same name exists (default is to prompt)")
    parser.add_option("-o", action="store_const", const="always", dest="overwrite",
                      help="Always overwrite files.  If a file exists in the destination "
                      "with the same name, it will be overwritten by the archive member "
                      "(default is to prompt)")
    parser.set_defaults(overwrite="prompt")
    (options, args) = parser.parse_args()
    if len(args) == 0:
        parser.print_help()
        sys.exit()
    
    #Process any wildcards and construct the list of zipfiles
    if (args[0].find('*') != -1 or
        args[0].find('?') != -1 or
        args[0].find('[') != -1):
        filelist = []
        (path, pattern) = os.path.split(args[0])
        if not path:
            searchpath = '.'
        else:
            searchpath = path
        for file in os.listdir(searchpath):
            if fnmatch.fnmatch(file, pattern):
                filelist.append(os.path.join(path, file))
    else:
        filelist = (args[0],)
        
    whitelist = []
    if len(args) > 1:
        whitelist = args[1:]
            
    for file in filelist:
        #Append the .zip suffix to the filename if necessary
        if not zipfile.is_zipfile(file):
            if zipfile.is_zipfile(file+".zip"):
                file += ".zip"
            else:
                print file, "is not a valid zipfile, skipping"
                continue
        
        print "Archive:  ", file
        zip = zipfile.ZipFile(file, "r", allowZip64=True)
        if options.display_comment:
            list_comment(zip)
        else:
            contents = [ x for x in zip.infolist() if (pass_whitelist(x.filename, whitelist) and
                                                       pass_blacklist(x.filename, options.exclude))]
            if options.list:
                list_files(zip, contents)
            elif options.test:
                test_files(zip, contents)
            else:
                for x in contents:
                    x.filename = adjust_path(x.filename, options.root, options.junk)
                extract(zip, contents, options.overwrite)
        zip.close()
        #Blank line between archives, if multiple are specified
        if len(filelist) > 1:
            print
    if len(filelist) > 1:
        print "{0} archives were successfully processed.".format(len(filelist))